@@ -16,7 +16,7 @@ from typing_extensions import Annotated
 import napari
 import numpy as np
 
-from .._categories import Category, find_function
+from .._categories import Category, find_function, get_name_of_function
 from qtpy.QtWidgets import QPushButton, QDockWidget
 
 try:
@@ -32,6 +32,7 @@ if TYPE_CHECKING:
 VIEWER_PARAM = "viewer"
 OP_NAME_PARAM = "op_name"
 OP_ID = "op_id"
+DEFAULT_BUTTON_SIZE = 40
 
 # We currently support operations with up to 6 numeric parameters, 3 booleans and 3 strings (see lists below)
 FloatRange = Annotated[float, {"min": np.finfo(np.float32).min, "max": np.finfo(np.float32).max}]
@@ -67,6 +68,55 @@ def num_positional_args(func, types=[np.ndarray, napari.types.ImageData, napari.
     params = signature(func).parameters
     return len([p for p in params.values() if p.annotation in types])
 
+def kwarg_key_adapter(func):
+    '''
+    Returns a dictionary mapping keywords between the category kwarg keys used
+    by the napari assistant to the kwarg keys of the input functions as well
+    as the reverse mapping.
+
+    Parameters
+    ----------
+
+    func: function
+        the function for which the mapping should be generated
+    '''
+    adapter = {}
+    sig = signature(func)
+    items = sig.parameters.items()
+    # get the names of positional parameters in the new operation
+    param_names, numeric_param_names, bool_param_names, str_param_names = separate_argnames_by_type(
+        items)
+
+    # go through all parameters and collect their values in an args-array
+    num_count = 0
+    str_count = 0
+    bool_count = 0
+    for key in param_names: 
+        if key in numeric_param_names:
+            adapter[category_args_numeric[num_count]] = key
+            adapter[key] = category_args_numeric[num_count]
+            num_count += 1
+        elif key in bool_param_names:
+            adapter[category_args_bool[bool_count]] = key
+            adapter[key] = category_args_bool[bool_count]
+            bool_count += 1
+        elif key in str_param_names:
+            adapter[category_args_text[str_count]] = key
+            adapter[key] = category_args_text[str_count]
+            str_count += 1
+
+    other_count = 0    
+    other_param_names = [
+        name
+        for name, param in items
+        if param.annotation not in {int, str, float, bool}
+    ]
+    for key in other_param_names:
+        adapter["input" + str(other_count)] = key
+        adapter[key] = "input" + str(other_count)
+        other_count += 1
+
+    return adapter
 
 @logger.catch
 def call_op(op_name: str, inputs: Sequence[Layer], timepoint : int = None, viewer: napari.Viewer = None, **kwargs) -> np.ndarray:
@@ -90,44 +140,20 @@ def call_op(op_name: str, inputs: Sequence[Layer], timepoint : int = None, viewe
     if not inputs or inputs[0] is None:
         return
 
-    # transfer data to gpu
-    if timepoint is None:
-        i0 = inputs[0].data
-        gpu_ins = [i.data if i is not None else i0 for i in inputs]
-    else:
-        i0 = inputs[0].data[timepoint] if len(inputs[0].data.shape) == 4 else inputs[0].data
-        gpu_ins = [(i.data[timepoint] if len(i.data.shape) == 4 else i.data if i is not None else i0) for i in inputs]
-
-    # convert 3d-1-slice-data into 2d data
-    # to support 2d timelapse data
-    gpu_ins = [i if len(i.shape) != 3 or i.shape[0] != 1 else i [0] for i in gpu_ins]
+    i0 = inputs[0].data
+    gpu_ins = [i.data if i is not None else i0 for i in inputs]
 
     # call actual cle function ignoring extra positional args
     cle_function = find_function(op_name)
     nargs = num_positional_args(cle_function)
 
-    args = []
     new_sig = signature(cle_function)
+    adapter = kwarg_key_adapter(cle_function)
     # get the names of positional parameters in the new operation
-    param_names, numeric_param_names, bool_param_names, str_param_names = separate_argnames_by_type(
+    param_names, _, _, _ = separate_argnames_by_type(
         new_sig.parameters.items())
-
-    # go through all parameters and collect their values in an args-array
-    num_count = 0
-    str_count = 0
-    bool_count = 0
-    for key in param_names:
-        if key in numeric_param_names:
-            value = kwargs[category_args_numeric[num_count]]
-            num_count = num_count + 1
-        elif key in bool_param_names:
-            value = kwargs[category_args_bool[bool_count]]
-            bool_count = bool_count + 1
-        elif key in str_param_names:
-            value = kwargs[category_args_text[str_count]]
-            str_count = str_count + 1
-        args.append(value)
-    args = tuple(args)
+    
+    args = tuple([kwargs[adapter[key]] for key in param_names])
 
     if cle_function.__module__ == "pyclesperanto_prototype":
         # todo: we should handle all functions equally
@@ -135,7 +161,8 @@ def call_op(op_name: str, inputs: Sequence[Layer], timepoint : int = None, viewe
 
         logger.info(f"{op_name}(..., {', '.join(map(str, args))})")
         args = ((*gpu_ins, gpu_out) + args)[:nargs]
-        gpu_out = cle_function(*args)
+
+        gpu_out = cle_function(*args, viewer=viewer)
 
         # return output
         return gpu_out, args
@@ -220,13 +247,16 @@ def _show_result(
         logger.warning("no viewer, cannot add image")
         return
     # show result in napari
-    clims = [gpu_out.min(), gpu_out.max()]
-
-    if clims[1] == 0:
-        clims[1] = 1
-
-    if clims[0] == clims[1]:
+    if "dask" in str(type(gpu_out)):
         clims = None
+    else:
+        clims = [gpu_out.min(), gpu_out.max()]
+
+        if clims[1] == 0:
+            clims[1] = 1
+
+        if clims[0] == clims[1]:
+            clims = None
 
     # conversion will be done inside napari. We can continue working with the potentially OCL-array from here.
     data = gpu_out
@@ -234,13 +264,13 @@ def _show_result(
     try:
         # look for an existing layer
         layer = next(x for x in viewer.layers if isinstance(x.metadata, dict) and x.metadata.get(OP_ID) == op_id)
-        logger.debug(f"updating existing layer: {layer}, with id: {op_id}")
+        # logger.debug(f"updating existing layer: {layer}, with id: {op_id}")
         layer.data = data
         layer.name = name
         # layer.translate = translate
     except StopIteration:
         # otherwise create a new one
-        logger.debug(f"creating new layer for id: {op_id}")
+        # logger.debug(f"creating new layer for id: {op_id}")
         add_layer = getattr(viewer, f"add_{layer_type}")
         kwargs = dict(name=name, metadata={OP_ID: op_id})
         if layer_type == "image":
@@ -251,17 +281,21 @@ def _show_result(
         layer = add_layer(data, **kwargs)
 
     if scale is not None:
-        if len(layer.data.shape) <= len(scale):
+        if len(layer.data.shape) == len(scale):
+            layer.scale = scale
+        if len(layer.data.shape) < len(scale):
             layer.scale = scale[-len(layer.data.shape):]
     return layer
 
 
-def _generate_signature_for_category(category: Category, search_string:str= None) -> Signature:
+def _generate_signature_for_category(category: Category, search_string:str= None, viewer:napari.Viewer = None) -> Signature:
     """Create an inspect.Signature object representing a cle Category.
 
     The output of this function can be used to set function.__signature__ so that
     magicgui can convert it into the appropriate widget.
     """
+    search_string = search_string.lower()
+
     k = Parameter.KEYWORD_ONLY
 
     # add inputs (we name them inputN ...)
@@ -294,14 +328,33 @@ def _generate_signature_for_category(category: Category, search_string:str= None
 
     # add a viewer.  This allows our widget to know if it's in a viewer
     params.append(
-        Parameter(VIEWER_PARAM, k, annotation="napari.viewer.Viewer", default=None)
+        Parameter(VIEWER_PARAM, k, annotation="napari.viewer.Viewer", default=viewer)
     )
     result = Signature(params)
     #print("Signature", result)
     return result
 
+def _get_operation_name_for_category_clicked(category: Category, search_string:str= None) -> str:
+    """Create an inspect.Signature object representing a cle Category.
 
-def make_gui_for_category(category: Category, search_string:str = None, viewer: napari.Viewer = None, button_size=40) -> magicgui.widgets.FunctionGui[Layer]:
+    The output of this function is the operation name for a clicked cataegory given a search string
+    """
+    # Add valid operations choices (will create the combo box)
+    from .._categories import operations_in_menu
+    choices = list(operations_in_menu(category, search_string))
+
+    default_op = category.default_op
+    if not any(default_op == op for op in choices):
+        #print("Default-operation is not in list!")
+        default_op = None
+
+    if default_op is None:
+        return choices[0]
+        
+    else:
+        return default_op
+
+def make_gui_for_category(category: Category, search_string:str = None, viewer: napari.Viewer = None, button_size=DEFAULT_BUTTON_SIZE, operation_name:str=None, autocall:bool=True) -> magicgui.widgets.FunctionGui[Layer]:
     """Generate a magicgui widget for a Category object
 
     Parameters
@@ -323,7 +376,15 @@ def make_gui_for_category(category: Category, search_string:str = None, viewer:
         We modify it's __signature__ below.
         """
         viewer = kwargs.pop(VIEWER_PARAM, None)
-        inputs = [kwargs.pop(k) for k in list(kwargs) if k.startswith("input")]
+        inputs = []
+        for i in [kwargs.pop(k) for k in list(kwargs) if k.startswith("input")]:
+            if isinstance(i, napari.layers.Layer):
+                inputs.append(i)
+            else:
+                from napari_workflows._workflow import _get_layer_from_data
+                inputs.append(_get_layer_from_data(viewer, i))
+
+        result_layer = None
         t_position = None
         if viewer is not None and len(viewer.dims.current_step) == 4:
             # in case we process a 4D-data set, we need read out the current timepoint
@@ -333,8 +394,12 @@ def make_gui_for_category(category: Category, search_string:str = None, viewer:
             currstep_event = viewer.dims.events.current_step
 
             def update(event):
-                currstep_event.disconnect(update)
-                widget()
+                #currstep_event.disconnect(update)
+                if result_layer is not None:
+                    from napari_workflows import WorkflowManager
+                    manager = WorkflowManager.install(viewer)
+                    manager.invalidate([result_layer.name])
+                #widget()
 
             if hasattr(widget, 'updater'):
                 currstep_event.disconnect(widget.updater)
@@ -379,7 +444,22 @@ def make_gui_for_category(category: Category, search_string:str = None, viewer:
             try:
                 from napari_workflows import WorkflowManager
                 manager = WorkflowManager.install(viewer)
-                manager.update(result_layer, find_function(op_name), *used_args)
+
+                # this step basically separates actual arguments from kwargs as this can cause 
+                # conflicts when setting the workflow step. 
+                signat = signature(find_function(op_name))
+                only_args = [
+                    arg for arg, param in zip(used_args,signat.parameters.values()) 
+                    if param.default is param.empty
+                ]
+                determined_kwargs = {
+                    name:value for (name,param),value in zip(signat.parameters.items(),used_args) 
+                    if not (param.default is param.empty)
+                }
+                # debugging prints
+                #print(f'only arguments: {only_args}')
+                #print(f'det kwargs:     {determined_kwargs}')
+                manager.update(result_layer, find_function(op_name), *only_args, **determined_kwargs)
                 #print("notified", result_layer.name, find_function(op_name))
             except ImportError:
                 pass # recording workflows in the WorkflowManager is a nice-to-have at the moment.
@@ -389,6 +469,8 @@ def make_gui_for_category(category: Category, search_string:str = None, viewer:
                 if layer in inputs or layer is result_layer:
                     try:
                         viewer.window.remove_dock_widget(widget.native)
+                    # TODO find more elegant or specific solution 
+                    # because this could break some stuff
                     except:
                         pass
 
@@ -398,16 +480,20 @@ def make_gui_for_category(category: Category, search_string:str = None, viewer:
         return None
 
     gui_function.__name__ = f'do_{category.name.lower().replace(" ", "_")}'
-    gui_function.__signature__ = _generate_signature_for_category(category, search_string)
+    gui_function.__signature__ = _generate_signature_for_category(category, search_string, viewer)
 
     # create the widget
-    widget = magicgui(gui_function, auto_call=category.auto_call)
+    widget = magicgui(gui_function, auto_call=autocall)
     widget.native.setMinimumWidth(100)
     modify_layout(widget.native, button_size=button_size)
 
+    if operation_name == None:
+        operation_name = _get_operation_name_for_category_clicked(category,search_string)
+    
     # when the operation name changes, we want to update the argument labels
     # to be appropriate for the corresponding cle operation.
     op_name_widget = getattr(widget, OP_NAME_PARAM)
+    op_name_widget.value = operation_name
 
     @op_name_widget.changed.connect
     def update_positional_labels(*_: Any):
